message "Creating lib DIRT, Please wait .."
createlib "dirt"
release procs wproc
 proc wproc(x)
    execute "writelib \"dirt\" " + x
    + " release procs " + x
 endproc

;================================================== DIRT_DirTree_Dlg
proc DIRT_DirTree_Dlg
; Dialog to select a directory from a tree.
; (Like {Tools} {More} {Directory} in version 4.5.
;
; Return Value:
; The selected directory name.
; Blank if no directory is selected (dialog cancelled).
(
    dir.name   ; {I} Initial directory name.
               ;     If len( dir.name ) <= 2, the dialog displays
               ;     the drive list. A default drive may be passed
               ;     as a drive specifier, i.e., "x:", where x is
               ;     a letter.
  , dlg.title  ; {I} The title for the dialog box
  , dlg.help   ; {I} The name of a proc to display help for the
               ;     dialog. The proc is executed via an EXECPROC.
               ;     If no help proc is provided, then a generic
               ;     help screen is used.
)
private x, y, z
    , butval
    , drive.list        ; Array of "x:", where x = drive letter.
                        ; Drive letters are in ascending order.
    , dir.tree          ; Array containing dir tree
    , dir.list          ; Dynarray containing components of selected
                        ; directory (in dir.name) and it's subdirs.
                        ; Indexed sequentially starting at 1.
    , dir.level         ; Index (in dir.list) of the rightmost
                        ; subdir in dir.name.
    , dir.tree.ndx      ; Index of selected element in dir.tree.
                        ; (Set by dialog when user pushes the
                        ; Select button and cursor is on the
                        ; directory tree.)
    , dir.drive         ; Drive specifier (x:)
    , dirtree.dir       ; The current directory expanded in the
                        ; directory tree.
    , dirtree.search.count  ; Counter for dir tree incremental search.
    , dirtree.search.string ; String for dir tree incremental search.


; Make sure the globals for utility procs are initialized.
if not isassigned( utlx.globals.are.initialized ) then
    UTLX_Init_Globals()
endif

dirtree.search.count = 0
dirtree.search.string  = ""
dirtree.dir = ""  ; no dir in tree yet

; Set default help
if isblank( dlg.help ) then
    dlg.help = "DIRT_DirTree_Dlg_Help"
endif

; Enable messages during initialization
utlx.direxists.msg = true
utlx.drivestatus.msg = true

dir.name = upper( dir.name )
dir.drive = substr( dir.name, 1, 2 )
if len( dir.name ) >= 3 then
    if substr( dir.name, len( dir.name ), 1 ) <> "\\" then
        dir.name = dir.name + "\\"
    endif
else
    ; dir.name contains the default drive not a dir name.
    dir.name = ""
endif

; Get the drive list and sort it
sysinfo drives to dir.list
array drive.list[ dynarraysize( dir.list ) ]
x = 0
foreach y in dir.list
    x = x + 1
    drive.list[ x ] = y
endforeach
release vars dir.list
for x from 1 to arraysize( drive.list )
    for y from x + 1 to arraysize( drive.list )
        if drive.list[ y ] < drive.list[ x ] then
            z = drive.list[ x ]
            drive.list[ x ] = drive.list[ y ]
            drive.list[ y ] = z
        endif
    endfor
endfor

; If no dir name was passed by caller, display the drive list.
; Otherwise, display the tree for the current directory and
; its immediate subdirs.
if isblank( dir.name ) then
    DIRT_DirTree_Dlg_Drives()
else
    DIRT_DirTree_Dlg_Tree()
endif

butval = "cancel"

SHOWDIALOG
    dlg.title
    PROC "DIRT_DirTree_DlgProc"
        TRIGGER "ALL"
;;      MOUSE "ALL"
        KEY "ALL"
;;      MESSAGE "ALL"
;;      IDLE
    @ 1,10
    HEIGHT 21
    WIDTH 59

    LABEL
        @ 1,2
        "~D~irectory"
        FOR "dir.name"

    ACCEPT
        @ 2,2
        WIDTH 53
        "A60"
        TAG "dir.name"
        TO dir.name

    LABEL
        @ 4,2
        "Directory ~T~ree"
        FOR "dir.tree.ndx"

    PICKARRAY
        @ 5,2
        HEIGHT  13
        WIDTH   39
        COLUMNS 1
        dir.tree
        TAG "dir.tree.ndx"
        TO dir.tree.ndx

    PUSHBUTTON
        @ 6,43
        WIDTH 10
        "~O~k"
        OK
        VALUE "ok"
        TAG "ok"
        TO butval

    PUSHBUTTON
        @ 8,43
        WIDTH 10
        "~S~elect"
        DEFAULT
        OK
        VALUE "select"
        TAG "select"
        TO butval

    PUSHBUTTON
        @ 10,43
        WIDTH 10
        "~C~reate"
        OK
        VALUE "create"
        TAG "create"
        TO butval

    PUSHBUTTON
        @ 12,43
        WIDTH 10
        "~H~elp"
        OK
        VALUE "help"
        TAG "help"
        TO butval

    PUSHBUTTON
        @ 14,43
        WIDTH 10
        "Cancel"
        Cancel
        VALUE "cancel"
        TAG "cancel"
        TO butval

ENDDIALOG

if lower( butval ) = "cancel" then
    return ""
else
    return dir.name
endif

endproc
wproc( "DIRT_DirTree_Dlg" )

;================================================== DIRT_DirTree_DlgProc
proc DIRT_DirTree_DlgProc
; Dialog proc for DIRT_DirTree_Dlg.
;
; Return Value:
; True if the event is to be processed.
; False if the event is to be ignored.
(
    dlg.trig     ; "Event" or one of the following triggers:
                 ; "Update", "Depart", "Select"
                 ; "Accept", "Cancel", "Arrive"
                 ; "Open", "Close"
  , dlg.tag      ; The tag for the active control element
  , dlg.evt      ; Dynarray for event.
                 ; If "Update", value of active control element.
                 ; If "Depart", tag that you will navigate to.
                 ; Null string if any other trigger.
  , dlg.cb.lbl   ; Null string if event.
                 ; If "Update" trigger, label string of checkbox or
                 ; null string.
                 ; Null string if any other trigger.
)
private x
switch
case dlg.trig = "EVENT" :
    switch
    case dlg.evt[ "type" ] = "IDLE"    : return false
    case dlg.evt[ "type" ] = "KEY"     : return DIRT_DirTree_DlgEvt_Keys()
    case dlg.evt[ "type" ] = "MOUSE"   : return true
    case dlg.evt[ "type" ] = "MESSAGE" : return true
    endswitch
case dlg.trig = "UPDATE" : return DIRT_DirTree_DlgTrg_Update()
case dlg.trig = "DEPART" : return DIRT_DirTree_DlgTrg_Depart()
case dlg.trig = "SELECT" : return DIRT_DirTree_DlgTrg_Select()
case dlg.trig = "ACCEPT" : return DIRT_DirTree_DlgTrg_Accept()
case dlg.trig = "ARRIVE" : return true
case dlg.trig = "OPEN"   : return DIRT_DirTree_DlgTrg_Open()
case dlg.trig = "CANCEL" : return true
case dlg.trig = "CLOSE"  : return true
endswitch
return true

endproc
wproc( "DIRT_DirTree_DlgProc" )

;============================================== DIRT_DirTree_DlgEvt_Keys
proc DIRT_DirTree_DlgEvt_Keys
; Handles KEY events for the directory tree dialog.
;
; Return Value:
; True or false.
(
)
private x, y, z
    , sp

; Increment counter for dir tree incremental search
dirtree.search.count = dirtree.search.count + 1

; Enable messages for this interaction
utlx.direxists.msg = true
utlx.drivestatus.msg = true

; Treat F2 as Alt-O (for the OK button)
if dlg.evt[ "keycode" ] = asc( "F2" ) then
    release vars dlg.evt[ "scancode" ]
    dlg.evt[ "keycode" ] = -24
    return true
endif

; It's not really necessary, but we'll kill the Ctrl-O key.
; We allow Alt-O because it is tied to the Ok button for the dialog.
if dlg.evt[ "keycode" ] = 15 then
    beep
    return false
endif

; Show help if the user presses F1
if dlg.evt[ "keycode" ] = asc( "F1" ) then
    execproc dlg.help
    return false
endif

; If the user presses Esc while in the dir name field and
; the value in the display field is shorter than the dir
; name, then redisplay with the dir name.
; When the cursor arrives on the dir name field, the entire field
; is selected. If the user types a char (intending to extend the
; field), the char replaces the entire field.
; Most of the time this select/replace behavior is desireable.
; However, there are times when it is a pain in the butt.
if lower( dlg.tag ) = "dir.name" and
   dlg.evt[ "keycode" ] = 27 then
    x = controlvalue( dlg.tag )
    if len( x ) < len( dir.name ) then
        refreshcontrol dlg.tag
        return false
    endif
    return true
endif

; We only care about ASCII keystrokes < 127 in the dir tree field.
if lower( dlg.tag ) = "dir.tree.ndx" and
   dlg.evt[ "keycode" ] > 32 and
   dlg.evt[ "keycode" ] < 127 then
    z = upper( chr( dlg.evt[ "keycode" ] ) )

    ; If the dir name is blank, then we are in the drive list.
    ; Only letters are used in the incremental search.
    if isblank( dir.name ) then
        if z >= "A" and z <= "Z" then
            for x from 1 to arraysize( drive.list )
                if z = substr( drive.list[ x ], 1, 1 ) then
                    dir.tree.ndx = x + 1
                    refreshcontrol "dir.tree.ndx"
                    return false
                endif
            endfor
        endif
        return true
    endif

    ; If the number of keystrokes since the last incremental
    ; search key is greater than 1, then reset the search field.
    if dirtree.search.count > 1 then
        dirtree.search.string = z
    else
        dirtree.search.string = dirtree.search.string + z
    endif
    ; If the search string is 1 byte long, start the search at the
    ; top of the tree. Otherwise, start at current entry.
    if len( dirtree.search.string ) = 1 then
        sp = 2    ; ignore drive id
    else
        sp = dir.tree.ndx - 1
    endif
    dirtree.search.count = 0        ; reset counter
    z = dirtree.search.string
    for x from sp to dynarraysize( dir.list )
        if z = substr( dir.list[ x ], 1, len(z) ) then
            dir.tree.ndx = x + 1
            refreshcontrol "dir.tree.ndx"
            return false
        endif
    endfor
    ; Search failed - Strip last character
    if len( z ) < 2 then
        dirtree.search.string = ""
    else
        dirtree.search.string = substr( z, 1, len( z ) - 1 )
    endif
    return true
endif

return true

endproc
wproc( "DIRT_DirTree_DlgEvt_Keys" )

;============================================== DIRT_DirTree_DlgTrg_Update
proc DIRT_DirTree_DlgTrg_Update
; Handles UPDATE trigger for the directory tree dialog.
;
; Return Value:
; True or false.
(
)
private x, y

return true

endproc
wproc( "DIRT_DirTree_DlgTrg_Update" )

;============================================== DIRT_DirTree_DlgTrg_Depart
proc DIRT_DirTree_DlgTrg_Depart
; Handles DEPART trigger for the directory tree dialog.
;
; Return Value:
; True or false.
(
)
private x, y, z

switch
case lower( dlg.evt ) = "cancel" :
    ; If heading to CANCEL, then allow it unconditionally.
    return true

case lower( dlg.evt ) = "help" :
    ; If the cursor is in the dir name or tree, then show
    ; help but stay in the field.
    if lower( dlg.tag ) = "dir.name" or
       lower( dlg.tag ) = "dir.tree.ndx" then
        execproc dlg.help
        return false
    endif

    ; If heading to HELP from any other field, then allow it
    ; unconditionally.
    return true

case isblank( dir.name ) and
     lower( dlg.tag ) = "dir.name" :
    ; If leaving dir name field and dir.name is blank, then
    ; show the drive list and position to dir tree field.
    DIRT_DirTree_Dlg_Drives()
    refreshcontrol( "dir.name" )
    refreshcontrol( "dir.tree.ndx" )
    selectcontrol "dir.tree.ndx"
    return false

case lower( dlg.tag ) = "dir.name" :
    ; If leaving the dir name field (user can enter dir name),
    ; then we must validate the dir name.
    ; We know that dir.name is not blank. (See previous case.)

    if dir.name = dirtree.dir then
        ; This is the same dir that is already in the
        ; dir tree. We won't check it. There is some
        ; chance that the user removed the disk from the
        ; drive, but the probability is low. Anyway, we'll
        ; catch the error later.
        return true
    endif

    if lower( dlg.evt ) = "create" then
        ; We are heading to the Create button. Let it happen.
        return true
    endif

    x = dir.name   ; use temp var (don't modify dir.name yet)
    ; Enforce upper case and make sure dir.name ends with \
    x = upper( x )
    if substr( x, len( x ), 1 ) <> "\\" then
        x = x + "\\"
    endif
    ; Make sure the drive is valid and ready
    if not DIRT_DirTree_Dlg_IsDriveOk( x ) then
        return false
    endif
    ; Make sure the dir exists
    if not UTLX_DirExists( x ) then
        UTLX_Show_Error( "Directory \"" + x + "\" does not exist." )
        selectcontrol dlg.tag
        return false
    endif
    dir.name = x     ; Set dir.name to new value
    ; If we are heading to the OK button, then let it happen.
    ; Otherwise, rebuild the tree and position to the dir tree
    ; field.
    if lower( dlg.evt ) = "ok" then
        return true
    endif
    DIRT_DirTree_Dlg_Tree()
    refreshcontrol( "dir.name" )
    refreshcontrol( "dir.tree.ndx" )
;;    refreshdialog
    selectcontrol "dir.tree.ndx"
    return false

endswitch

return true

endproc
wproc( "DIRT_DirTree_DlgTrg_Depart" )

;============================================== DIRT_DirTree_DlgTrg_Select
proc DIRT_DirTree_DlgTrg_Select
; Handles SELECT trigger for the directory tree dialog.
;
; Return Value:
; True or false.
(
)
private x

switch
case lower( dlg.tag ) = "dir.tree.ndx" and dir.tree.ndx = 1 :
    ; If we are in the dir tree picklist and the user selected
    ; the first line, then we want to display the drive list
    ; in the dir tree picklist.
    dir.name = ""   ; no directory selected
    DIRT_DirTree_Dlg_Drives()
    refreshcontrol( "dir.name" )
    refreshcontrol( "dir.tree.ndx" )
;;    refreshdialog

case lower( dlg.tag ) = "dir.tree.ndx" and isblank( dir.name ) :
    ; If we are in the dir tree picklist and there is no
    ; current dir name, then the picklist contains the drive list.

    ; Make sure the selected drive is valid and ready
    x = drive.list[ dir.tree.ndx - 1 ]
    if not DIRT_DirTree_Dlg_IsDriveOk( x ) then
        selectcontrol dlg.tag
        return false
    endif

    ; Selected dir is root of drive. Show tree for dir.
    dir.name = x + "\\"
    DIRT_DirTree_Dlg_Tree()
    refreshcontrol( "dir.name" )
    refreshcontrol( "dir.tree.ndx" )
;;    refreshdialog

case lower( dlg.tag ) = "dir.tree.ndx" :
    ; We are in the dir tree picklist and it contains the tree
    ; for dir.name

    switch
    case dir.tree.ndx - 1 < dir.level :
        ; The selected entry is a parent of dir.name.
        dir.name = UTLX_Concat_Path( dir.list, dir.tree.ndx - 1 )

    case dir.tree.ndx - 1 > dir.level :
        ; The selected entry is a subdir of dir.name.
        ; Concatenate it to dir.name
        dir.name = dir.name + dir.list[ dir.tree.ndx - 1 ] + "\\"

    otherwise :
        ; The user selected the same dir again. We don't have to
        ; do anything special.

    endswitch

    ; Rebuild the dir tree and redisplay the dialog
    DIRT_DirTree_Dlg_Tree()
    refreshcontrol( "dir.name" )
    refreshcontrol( "dir.tree.ndx" )
;;    refreshdialog

endswitch

return true

endproc
wproc( "DIRT_DirTree_DlgTrg_Select" )

;============================================== DIRT_DirTree_DlgTrg_Accept
proc DIRT_DirTree_DlgTrg_Accept
; Handles ACCEPT trigger for the directory tree dialog.
;
; Return Value:
; True or false.
(
)
private x

switch
case lower( dlg.tag ) = "ok" and isblank( dir.name ) :
    ; If the user tries to leave without selecting a dir,
    ; then issue a message and show the drive list.
    UTLX_Show_Error( "No directory has been selected!" )
    DIRT_DirTree_Dlg_Drives()
    refreshcontrol( "dir.name" )
    refreshcontrol( "dir.tree.ndx" )
    selectcontrol "dir.tree.ndx"
    return false

case lower( dlg.tag ) = "help" :
    ; If user pushes the HELP button, then show the help and
    ; cancel the accept.
    execproc dlg.help
    return false

case lower( dlg.tag ) = "create" :
    ; If user pushes the CREATE button, then create the named dir
    ; and reject the accept.
    if isblank( dir.name ) then
        beep
        return false
    endif
    if UTLX_Make_Dir( dir.name ) then
        ; Dir was created - Rebuild tree and reshow dialog.
        if substr( dir.name, len( dir.name ), 1 ) = "\\" then
            dir.name = upper( dir.name )
        else
            dir.name = upper( dir.name + "\\" )
        endif
        DIRT_DirTree_Dlg_Tree()
        refreshcontrol( "dir.name" )
        refreshcontrol( "dir.tree.ndx" )
        selectcontrol "dir.tree.ndx"
    endif
    return false

case lower( dlg.tag ) = "select" :
    ; If user pushes the SELECT button, then call the dialog proc
    ; with a fake SELECT trigger. Then position to the tree and
    ; reject the accept.
    ; (The SELECT trigger will refresh the dialog and position to
    ; the dir tree.)
    DIRT_DirTree_DlgProc( "SELECT", "dir.tree.ndx" , "", "" )
    selectcontrol "dir.tree.ndx"
    return false

endswitch

return true

endproc
wproc( "DIRT_DirTree_DlgTrg_Accept" )

;============================================== DIRT_DirTree_DlgTrg_Open
proc DIRT_DirTree_DlgTrg_Open
; Handles OPEN trigger for the directory tree dialog.
;
; Return Value:
; True or false.
(
)
private x

; If no dir name, then position to dir tree picklist.
; The picklist will contain the drive list.
if isblank( dir.name ) then
    selectcontrol "dir.tree.ndx"
endif

return true

endproc
wproc( "DIRT_DirTree_DlgTrg_Open" )

;============================================ DIRT_DirTree_Dlg_IsDriveOk
proc DIRT_DirTree_Dlg_IsDriveOk
; Check the drive for the directory tree dialog.
; Uses vars defined in the dialog and associated procs.
;
; Checks the following:
; - Is it a valid drive specification? (x: where A <= x <= Z)
; - Does the drive exist? (Is it in the drive table?)
; - Is the drive ready?
; - Is the disk formatted?
;
; If an invalid condition is detected, an error is issued.
;
; Return Value:
; True if the drive passes all tests.
; False if the drive fails any test.
(
    drv       ; {I} Two byte drive spec - x:, where x is a letter.
              ;     If more than two bytes are passed, only the
              ;     first two bytes are considered. I.e, you can
              ;     pass a full directory string and only the drive
              ;     will be validated.
)
private x
    , drv.let
    , drv.sep

drv.let = upper( substr( drv, 1, 1 ) )
drv.sep = substr( drv, 2, 1 )

drv = drv.let + drv.sep

if drv.sep <> ":" or
   drv.let < "A" or drv.let > "Z" then
    UTLX_Show_Error( "\"" + drv + "\" is not a valid drive "
            + "specification." )
    return false
endif

for x from 1 to arraysize( drive.list )
    if drv = drive.list[ x ] then
        quitloop
    endif
endfor
if x > arraysize( drive.list ) then
    UTLX_Show_Error( "\"" + drv + "\" is not an attached drive." )
    return false
endif

if not UTLX_DriveStatus( drv ) then
    UTLX_Show_Error( "Drive \"" + drv + "\" is not ready.\n\n"
        + "Please insert a formatted disk in the drive\n"
        + "and close the drive door/latch." )
    utlx.drivestatus.msg = true       ; enable the message
    if not UTLX_DriveStatus( drv ) then        ; try again
        return false
    endif
endif

; Check for root directory. If it does not exist, the disk is not
; formatted.
; Note: The following test will sometimes succeed even if the
; disk is unformatted. This seems to occur (sometimes) when
; a formatted disk is inserted and checked and then an unformatted
; disk is checked.
if UTLX_DriveStatus( drv ) then
    if not UTLX_DirExists( drv + "\\" ) then
        UTLX_Show_Error( "The disk in drive \""
        + drv + "\" is not formatted." )
        return false
    endif
else
    return false   ; somebody's pretty quick at removing disks
endif

return true    ; drive is ok

endproc
wproc( "DIRT_DirTree_Dlg_IsDriveOk" )

;============================================= DIRT_DirTree_Dlg_Init_Tree
proc DIRT_DirTree_Dlg_Init_Tree
; Init the directory tree array used by the dialog in proc
; DIRT_DirTree_Dlg
;
; Return Value:
; None.
; The array dir.tree is allocated and initialized.
(
    array.size   ; {I} Size to allocate.
)
private x

array dir.tree[ array.size ]
for x from 2 to array.size
    dir.tree[ x ] = ""
endfor
dir.tree[ 1 ] = "Drives"

endproc
wproc( "DIRT_DirTree_Dlg_Init_Tree" )

;================================================== DIRT_DirTree_Dlg_Drives
proc DIRT_DirTree_Dlg_Drives
; Build drive list for the dialog in proc DIRT_DirTree_Dlg.
; Uses vars defined in the above named proc.
;
; Return Value:
; None.
(
)
private x

DIRT_DirTree_Dlg_Init_Tree( arraysize( drive.list ) + 1 )

dir.tree.ndx = 0
for x from 1 to arraysize( drive.list )
    ; Set index for default drive
    if dir.drive = drive.list[ x ] then
        dir.tree.ndx = x + 1
    endif
    ; Note:  = chr(195),  = chr(192), and  = chr(196)
    if x = arraysize( drive.list ) then
        dir.tree[ x + 1 ] = " " + drive.list[ x ]
    else
        dir.tree[ x + 1 ] = " " + drive.list[ x ]
    endif
endfor
if dir.tree.ndx = 0 then
    dir.drive = ""        ; default drive is not in drive list
endif
dirtree.dir = ""  ; No current dir for tree

endproc
wproc( "DIRT_DirTree_Dlg_Drives" )

;================================================== DIRT_DirTree_Dlg_Tree
proc DIRT_DirTree_Dlg_Tree
; Build directory tree for the dialog in proc DIRT_DirTree_Dlg.
; Uses vars defined in the above named proc.
;
; Directory name is in dir.name.
;
; If the drive is not ready or the dir in dir.name doesn't exist,
; then the drive list will be displayed.
;
; Return Value:
; None.
; The directory tree or drive list will be built in the
; dynarray dir.tree. The var dir.tree.ndx will be set to
; the current entry in dir.tree.
(
)
private x, y, z
    , subdir.name

if isblank( dir.name ) then
    DIRT_DirTree_Dlg_Drives()  ; show drive list
else
    dir.drive = substr( dir.name, 1, 2 )  ; set default drive
    switch
    case dirtree.dir = dir.name :
        ; dir.name = current dir in tree (no need to rebuild tree)
        dir.tree.ndx = dir.level + 1
    case not UTLX_DriveStatus( dir.drive ) :
        ; If the drive is not ready, then show the drive list.
        dir.name = ""   ; no dir selected
        DIRT_DirTree_Dlg_Drives()
    case not UTLX_DirExists( dir.name ) :
        ; If the dir doesn't exist, then show the drive list.
        dir.name = ""   ; no dir selected
        DIRT_DirTree_Dlg_Drives()
    otherwise :
        dynarray dir.list[]
        UTLX_Parse_Path( dir.name, dir.list )
        dir.level = dynarraysize( dir.list )  ; lowest subdir in dir.name
        dynarray subdir.name[]
        ; Add subdirs of dir.name to dir.list
        DIRT_Get_Subdirs( dir.name, subdir.name )
        for x from 1 to dynarraysize( subdir.name )
            dir.list[ x + dir.level ] = subdir.name[x]
        endfor

        ; Initialize and draw the tree ( dir.tree[1] = "Drives" )
        DIRT_DirTree_Dlg_Init_Tree( dynarraysize( dir.list ) + 1 )
        for x from 1 to dynarraysize( dir.list )
            ; Note:  = chr(192),  = chr(195), and  = chr(196)
            switch
            case x <= dir.level :
                ; For drive and subdirs in dir.name
                z = spaces( 2 * x - 1 ) + ""
                dir.tree[ x + 1 ] =  z + dir.list[ x ]
            case x < dynarraysize( dir.list ) :
                ; For subdirs of dir.name (except last subdir)
                z = spaces( 2 * dir.level + 1 ) + ""
                dir.tree[ x + 1 ] = z + dir.list[ x ]
            otherwise :
                ; For last subdir of dir.name
                z = spaces( 2 * dir.level + 1 ) + ""
                dir.tree[ x + 1 ] = z + dir.list[ x ]
            endswitch
        endfor
        dir.tree.ndx = dir.level + 1
        dirtree.dir = dir.name  ; current dir in tree
    endswitch
endif

endproc
wproc( "DIRT_DirTree_Dlg_Tree" )

;============================================== DIRT_DirTree_Dlg_Help
proc DIRT_DirTree_Dlg_Help
; Generic help screen for the directory tree dialog proc.
;
; Return Value:
; None.
(
)
private x, y, z
    , orig.win
    , orig.canv
    , msgw
    , msgx
    , num.lines
    , widest.line
    , msg.hgt
    , msg.wid
    , msg.row
    , msg.col
    , msg.winhand
    , msg.can.wid
    , msg.can.hgt
    , scroll.col
    , scroll.row
    , win.attr

orig.win = getwindow()
orig.canv = getcanvas()

fileread sdir() + "dtreedlg.txt" to msgx
dynarray msgw[]
msgw[ "string" ] = msgx
release vars msgx
msgw[ "width" ] = 76
UTLX_Word_Wrap( msgw )

; Calc window sizes and locations. Pretty dry stuff.

num.lines = min( msgw[ "lines" ], 200 )
widest.line = msgw[ "longest" ]  ; won't exceed msgw[ "width" ]

msg.can.wid = 78
msg.can.hgt = num.lines
msg.can.hgt = min( msg.can.hgt, 200 )

msg.wid = 80
msg.hgt = 24

msg.row = 0
msg.col = 0

; Create the floating message window

; By the way - this demo assumes a screen size of 24x80.
; We could use sysinfo to get the actual size but this
; is only a demo - not a real application.

dynarray x[]
x[ "canclose" ] = false
x[ "canmaximize" ] = false
x[ "canmove" ] = false
x[ "canresize" ] = false
x[ "canvasheight"] = msg.can.hgt
x[ "canvaswidth" ] = msg.can.wid
x[ "floating" ] = true
x[ "hasframe" ] = true
x[ "hasshadow" ] = false
x[ "height" ] = msg.hgt
x[ "maximized" ] = false
x[ "origincol" ] = msg.col
x[ "originrow" ] = msg.row
x[ "width" ] = msg.wid

window create attributes x to msg.winhand
setcanvas msg.winhand
release vars x

dynarray x[]
x[ 0 ] = syscolor( 1032 )  ; use active color
x[ 1 ] = syscolor( 1032 )
x[ 2 ] = syscolor( 1032 )
x[ 3 ] = syscolor( 1034 )
x[ 4 ] = syscolor( 1035 )

window setcolors msg.winhand from x

; Put the message in the window and paint the entire window
; using the dialog text color.

for x from 1 to num.lines
    @ x - 1, 1 ?? msgw[ x ]
endfor

paintcanvas attribute syscolor( 1036 ) all

; Wait for the user to press Esc.

echo normal
prompt "Press Esc to Exit Help"

while true
    getevent key "all" mouse "all" to evt
    switch
    case evt[ "type" ] = "KEY" :
        if evt[ "keycode" ] = 27 then
            quitloop
        endif
        ;
        ;   Scrolling keys
        ;
        scroll.row = 0
        scroll.col = 0

        switch
        case evt[ "keycode" ] = -71 :  ; Home
            scroll.col = -1000         ; Scroll to left of canvas

        case evt[ "keycode" ] = -72 :  ; Up
            scroll.row = -1

        case evt[ "keycode" ] = -73 :  ; PgUp
            scroll.row = -msg.hgt - 1  ; 1 less than height

        case evt[ "keycode" ] = -75 :  ; Left
            scroll.col = -1

        case evt[ "keycode" ] = -77 :  ; Right
            scroll.col = 1

        case evt[ "keycode" ] = -79 :  ; End
            scroll.col = 1000          ; Scroll to right end of canvas

        case evt[ "keycode" ] = -80 :  ; Down
            scroll.row = 1

        case evt[ "keycode" ] = -81 :  ; PgDn
            scroll.row = msg.hgt - 3   ; 1 less than height

        case evt[ "keycode" ] = -115 : ; Ctrl-Left
            scroll.col = -msg.wid - 1  ; 1 less than width

        case evt[ "keycode" ] = -116 : ; Ctrl-Right
            scroll.col = msg.wid - 3   ; 1 less than width

        case evt[ "keycode" ] = -117 : ; Ctrl-End
            scroll.row = 1000          ; Scroll to bottom of canvas
            scroll.col = 1000          ; Scroll to right end of canvas

        case evt[ "keycode" ] = -118 : ; Ctrl-PgDn
            scroll.row = 1000          ; Scroll to bottom of canvas

        case evt[ "keycode" ] = -119 : ; Ctrl-Home
            scroll.row = -1000         ; Scroll to top of canvas
            scroll.col = -1000         ; Scroll to left end of canvas

        case evt[ "keycode" ] = -132 : ; Ctrl-PgUp
            scroll.row = -1000         ; Scroll to top of canvas

        otherwise :
            beep

        endswitch

        window getattributes msg.winhand to win.attr
        scroll.row = scroll.row + win.attr[ "scrollrow" ]
        scroll.row = max( scroll.row, 0 )
        scroll.row = min( scroll.row
            , win.attr[ "canvasheight" ] - msg.hgt )
        scroll.row = max( scroll.row, 0 )
        scroll.col = scroll.col + win.attr[ "scrollcol" ]
        scroll.col = max( scroll.col, 0 )
        scroll.col = min( scroll.col
            , win.attr[ "canvaswidth" ] - msg.wid )
        scroll.col = max( scroll.col, 0 )
        window scroll msg.winhand to scroll.row, scroll.col

    case evt[ "type" ] = "MOUSE" :
        localizeevent evt
        if evt[ "window" ] = msg.winhand then
            globalizeevent evt
            execevent evt
        endif
    endswitch
endwhile

; Kill the floating window

window select msg.winhand
window close

if iswindow( orig.canv ) then
    setcanvas orig.canv
else
    setcanvas default
endif

if iswindow( orig.win ) then
    window select orig.win
endif

prompt ""
echo normal echo off

endproc
wproc( "DIRT_DirTree_Dlg_Help" )

;================================================== DIRT_Get_Subdirs
proc DIRT_Get_Subdirs
; Get a list of subdirs for a given directory.
;
; Invokes a self-driven dialog with a pickfile to get
; the subdir list.
;
;
; Return Value:
; None.
; The list of subdirs is returned in dynarray subdir.name.
(
    dir.path         ; {I} Directory name ending in "\".
  , subdir.list      ; {O} Dynarray to hold subdir names.
                     ;     The dirs "." and ".." are not
                     ;     included in the list.
                     ;     The list is indexed starting at 1
                     ;     and is sorted in ascending sequence
                     ;     on dir name.
)
private x, y, z
    , picked.file
    , subdir.prev
    , subdir.ndx

; *.lck is used to minimize the number of files included in the list
dir.path = upper( dir.path + "*.lck" )
subdir.ndx = 0
subdir.prev = ""

pushkeys 32  ; force select trigger when dialog starts

SHOWDIALOG ""
    PROC "DIRT_Get_Subdirs_DlgProc"
        TRIGGER "SELECT"
        @ 100, 100
        HEIGHT 8
        WIDTH 28

    PickFile
        @ 1,3
        HEIGHT  3
        WIDTH   20
        COLUMNS 1
        dir.path
        showdirs
        TAG "picked.file"
    TO picked.file

ENDDIALOG

endproc
wproc( "DIRT_Get_Subdirs" )

;================================================== DIRT_Get_Subdirs_DlgProc
proc DIRT_Get_Subdirs_DlgProc
; Dialog proc for DIRT_Get_Subdirs
;
; Only the SELECT trigger is trapped.
;
; Return Value:
; True if the event is to be processed.
; False if the event is to be ignored.
(
    dlg.trig     ; "Event" or one of the following triggers:
                 ; "Update", "Depart", "Select"
                 ; "Accept", "Cancel", "Arrive"
                 ; "Open", "Close"
  , dlg.tag      ; The tag for the active control element
  , dlg.evt      ; Dynarray for event.
                 ; If "Update", value of active control element.
                 ; If "Depart", tag that you will navigate to.
                 ; Null string if any other trigger.
  , dlg.cb.lbl   ; Null string if event.
                 ; If "Update" trigger, label string of checkbox or
                 ; null string.
                 ; Null string if any other trigger.
)
private x, y, z

if dlg.trig = "SELECT" then   ;shouldn't be anything else
    x = controlvalue( "picked.file" )
    ; If we get the same one twice in a row, then we must be
    ; at the bottom of the pickfile list. All done.
    if x = subdir.prev then
        acceptdialog
    else
        pushkeys asc( "down" ), 32   ; select next
        subdir.prev = x
        ; Subdirs have \ at end of subdir name
        y = search( "\\", x )
        if y > 0 then
            x = upper( substr( x, 1, y - 1 ) )
            ; Don't include . or ..
            if substr( x, 1, 1 ) <> "." then
                subdir.ndx = subdir.ndx + 1
                subdir.list[ subdir.ndx ] = x
            endif
        endif
    endif
endif

return true

endproc
wproc( "DIRT_Get_Subdirs_DlgProc" )

;>
;> Utility procs to support the directory tree dialog.
;>
;> Some of these can probably be replaced by your own
;> equivalents.
;>
;> The utilities have the prefix "UTLX_"
;>
;> Global vars used by the utilities have the prefix "utlx."
;>
;\

;======================================================= UTLX_Init_Globals
proc UTLX_Init_Globals
; Initialize global Constants and vars for Utility (UTLX_*) procs
;
; Return value:
; None.
; Global constants and vars for utility procs are initialized.
(
)
;
; The following set the kind of sound produced for
; various types of messages.
;
utlx.beep.question = "UTLX_Double_Beep"
utlx.beep.error = "UTLX_Error_Beep"
utlx.beep.info = "UTLX_Single_Beep"
utlx.beep.interr = "UTLX_Crash_Beep"

; Debug mode can be set by setting the DOS environment
; variable named by utlx.debug.envir.sw to 1.
;
; The default name for the environment variable is
; PALDEBUG. To enable debug mode, use the following
; DOS command before starting Paradox:
;
;     SET PALDEBUG=1
;
; You can also set debug mode by setting utlx.debug.mode
; TRUE before calling this proc. If debug is already on,
; this proc won't look at the environment variable.

utlx.debug.envir.sw = "PALDEBUG"

if isassigned( utlx.debug.mode ) and
   utlx.debug.mode then
    ; If debug mode is already true, don't mess with it.
    ; This allows it to be set by the caller based on
    ; criteria other than the DOS environment variable.
else
    sysinfo environment to x
    if isassigned( x[ utlx.debug.envir.sw ] ) and
       x[ utlx.debug.envir.sw ] = "1" then
        utlx.debug.mode = true
    else
        utlx.debug.mode = false
    endif
    release vars x
endif


; Some utility code has TurboPAL dependencies.

sysinfo to x
utlx.turbopal = isassigned( x[ "turbopal" ] )
release vars x

;
; Enable messages (See procs UTLX_DirExists and UTLX_DriveStatus)
;
utlx.direxists.msg = true
utlx.drivestatus.msg = true

; Dummy field used to see if globals have been initialized.

utlx.globals.are.initialized = ""

endproc
wproc( "UTLX_Init_Globals" )

;===================================================== UTLX_Release_Globals
proc UTLX_Release_Globals
; Release the global Constants and vars for Utility (UTLX_*) procs
;
; Return value:
; None.
; Global constants and vars for utility procs are released.
(
)

release vars
      utlx.beep.question
    , utlx.beep.error
    , utlx.beep.info
    , utlx.beep.interr
    , utlx.debug.envir.sw
    , utlx.debug.mode
    , utlx.turbopal
    , utlx.direxists.msg
    , utlx.drivestatus.msg
    , utlx.globals.are.initialized

endproc
wproc( "UTLX_Release_Globals" )

;============================================== UTLX_Make_Dir
proc UTLX_Make_Dir
; Run DOS MKDIR command to make a directory.
;
; An entire path can be created in a single call.
; If the entire path cannot be created, then any
; subdirs that were created are removed.
;
; Return Value:
; True if the directory was created. Otherwise, false.
(
    new.dir       ; {I} String containing the dir to create.
                  ;     This is a full path starting at the
                  ;     root of the drive.
)
private x, y, z
    , path.part
    , first.mkdir
    , bad.mkdir

if isblank( new.dir ) then
    return false
endif

dynarray path.part[]
UTLX_Parse_Path( new.dir, path.part )

if isassigned( path.part[ 1 ] ) then
    y = path.part[ 1 ]
    if not UTLX_DriveStatus( y ) then
        UTLX_Show_Error( "Drive \"" + y + "\" is not ready." )
        return false
    endif
else
    beep
    return false
endif

first.mkdir = 0
for x from 2 to dynarraysize( path.part )
    y = UTLX_Concat_Path( path.part, x )
    if not UTLX_DirExists( y ) then
        first.mkdir = iif( first.mkdir = 0, x, first.mkdir )
        y = substr( y, 1, len( y ) - 1 )  ; remove \
        run norefresh "mkdir " + y
        if not UTLX_DirExists( y ) then
            bad.mkdir = x
            UTLX_Show_Error( "Cannot create directory \""
                + y + "\"\n\n"
                + iif( isfile( y )
                , "A file with the same name already exists."
                , "The directory name is invalid or there is not"
                + " enough space on the drive." ) )
            ; Remove any subdirs that were created
            if first.mkdir < bad.mkdir then
                for x from bad.mkdir - 1 to first.mkdir step -1
                    y = UTLX_Concat_Path( path.part, x )
                    y = substr( y, 1, len( y ) - 1 )  ; remove \
                    run norefresh "rmdir " + y
                endfor
            endif
            return false
        endif
    endif
endfor

return true

endproc
wproc( "UTLX_Make_Dir" )

;================================================== UTLX_DirExists
proc UTLX_DirExists
; Check the existence of a directory via a DIREXISTS function.
;
; Returns a boolean instead of an integer.
;
; Since DIREXISTS can take a long time to execute, a
; "Please Wait" message is displayed. Since this message will
; barely flicker for fast drives, a slight delay is added
; to make the message readable.
;
; Since this function may be called several times during an
; interaction, the message is automatically suppressed after
; the first call.
; To enable the message, the caller must set the global var
; utlx.direxists.msg to TRUE for the first call in a series.
;
; Return Value:
; True if DIREXISTS returns 1. Otherwise, false.
(
    dirname    ; {I} The directory to check.
)
private x
    , msghand

dirname = upper( dirname )

if not isassigned( utlx.direxists.msg ) then
    utlx.direxists.msg = true
endif

if utlx.direxists.msg then
    utlx.direxists.msg = false
    msghand = UTLX_Please_Wait( "Checking existence of directory:\n"
        + "\"" + dirname + "\"", 0, -8 )
    sleep 500
else
    msghand = 0
endif

x = direxists( dirname )

UTLX_Close_Canvas( msghand, 0 )

return ( x = 1 )

endproc
wproc( "UTLX_DirExists" )

;=========================================================== UTLX_Show_Error
proc UTLX_Show_Error
; Display a dialog box with an error message and an "Ok" button.
; (No help via F1)
;
; See UTLX_Show_ErrorH for an explanation of parameters.
; Return Value.
; None.
(
    err  ; {I} String containing the error message
)

UTLX_Show_ErrorH( err, "" )

endproc
wproc( "UTLX_Show_Error" )

;=========================================================== UTLX_Show_ErrorH
proc UTLX_Show_ErrorH
; Display a dialog box with an error message and an "Ok" button.
; (With Help via F1)
;
; Return Value.
; None.
(
    err        ; {I} The error message.
               ;     The message can be any length and will be
               ;     word wrapped.
  , help.proc  ; {I} Proc to execproc if the user presses F1.
               ;     If blank, F1 is ignored.
               ;     If not blank, the message should tell the
               ;     user that more help is available via F1.
)
private w, x, y, z

; Throw away any type-ahead characters
while charwaiting()
    x = getchar()
endwhile

execproc utlx.beep.error

UTLX_Ok_Box( err, "*** ERROR ***", help.proc, true, 79, "~O~k" )

endproc
wproc( "UTLX_Show_ErrorH" )

;=========================================================== UTLX_Ok_Box
proc UTLX_Ok_Box
; Display a dialog box with a message and a single button.
;
; Return Value:
; None.
(
    strng         ; {I} Message to display.
                  ;     The message will be word wrapped and can be
                  ;     any length. However, since it is stored in
                  ;     an array, a long message may use a lot of
                  ;     memory.
  , dlg.title     ; {I} Title for the dialog box.
  , dlg.help      ; {I} Proc to execproc if the user presses F1.
                  ;     If blank, F1 is ignored.
                  ;     If not blank, the message should tell the
                  ;     user that more help is available via F1.
  , centerit      ; {I} If TRUE, then each line of the message
                  ;     is centered.
  , msgcolor      ; {I} The attribute (color) for the text
  , but1          ; {I} The text for the single button.
                  ;     (max of 10 chars excluding ~)
                  ;     Hilight a character by enclosing between ~.
                  ;     Defaults to "~O~k"
)
private u, v, w, x, y, z
    , ww
    , msg
    , msghgt
    , msgwid
    , dlg.col
    , fmtw
    , butval
    , accept.butval

if isblank( but1 ) then
    but1 = "~O~k"
endif

butval = but1
accept.butval = but1

dynarray ww[]
ww[ "string" ] = strng
ww[ "width" ] = 74

UTLX_Word_Wrap( ww )

msghgt = ww[ "lines" ]
msgwid = max( ww[ "longest" ] + 6, 32 )    ; make sure it's at least 32 wide

x = len( dlg.title ) + 10
x = min( x, 80 )
msgwid = max( msgwid, x )

fmtw = "w" + strval( msgwid - 4 )
if centerit then
    fmtw = fmtw + ",ac"
endif

dlg.col = int( ( 80 - msgwid ) / 2 )

switch
case msghgt > 10 :
    array msg[ msghgt ]
    for x from 1 to msghgt
        msg[x] = format( fmtw, ww[x] )
    endfor
    x = msg[1]
    showdialog dlg.title
        PROC "UTLX_Box_DlgProc"
            TRIGGER "OPEN"
;;            MOUSE "ALL"
;;            KEY "ALL"
              KEY "F1", "F2"
;;            MESSAGE "ALL"
;;            IDLE
        @2, dlg.col
        height 19
        width msgwid

        pickarray
            @1,1
            height min( msghgt, 13 )
            width msgwid-4
            msg
            tag "msg"
            to x

        pushbutton
            @15,int( (msgwid - 17 ) / 2 )
            width 14
            but1
            cancel
            default
            value but1
            tag but1
            to butval

    enddialog

case msghgt > 5 :
    z = msghgt
    w = int( ( 10 - z ) / 2 )
    if w > 0 then
        for x from 10 to 1 step -1
            switch
            case z = 0 :
                ww[x] = ""
            case isassigned( ww[x-w] ) :
                z = z - 1
                ww[x] = ww[x-w]
            endswitch
        endfor
    endif

    showdialog dlg.title
        PROC "UTLX_Box_DlgProc"
            TRIGGER "OPEN"
;;            MOUSE "ALL"
;;            KEY "ALL"
              KEY "F1", "F2"
;;            MESSAGE "ALL"
;;            IDLE
        @4, dlg.col
        height 16
        width msgwid

        style attribute msgcolor
        @1,1  ?? format( fmtw, iif( isassigned( ww[1] ), ww[1], "" ) )
        @2,1  ?? format( fmtw, iif( isassigned( ww[2] ), ww[2], "" ) )
        @3,1  ?? format( fmtw, iif( isassigned( ww[3] ), ww[3], "" ) )
        @4,1  ?? format( fmtw, iif( isassigned( ww[4] ), ww[4], "" ) )
        @5,1  ?? format( fmtw, iif( isassigned( ww[5] ), ww[5], "" ) )
        @6,1  ?? format( fmtw, iif( isassigned( ww[6] ), ww[6], "" ) )
        @7,1  ?? format( fmtw, iif( isassigned( ww[7] ), ww[7], "" ) )
        @8,1  ?? format( fmtw, iif( isassigned( ww[8] ), ww[8], "" ) )
        @9,1  ?? format( fmtw, iif( isassigned( ww[9] ), ww[9], "" ) )
        @10,1 ?? format( fmtw, iif( isassigned( ww[10] ), ww[10], "" ) )
        style

        pushbutton
            @12,int( (msgwid - 17 ) / 2 )
            width 14
            but1
            cancel
            default
            value but1
            tag but1
            to butval

    enddialog

otherwise :
    z = msghgt
    w = int( ( 5 - z ) / 2 )
    if w > 0 then
        for x from 5 to 1 step -1
            switch
            case z = 0 :
                ww[x] = ""
            case isassigned( ww[x-w] ) :
                z = z - 1
                ww[x] = ww[x-w]
            endswitch
        endfor
    endif
    showdialog dlg.title
        PROC "UTLX_Box_DlgProc"
            TRIGGER "OPEN"
;;            MOUSE "ALL"
;;            KEY "ALL"
              KEY "F1", "F2"
;;            MESSAGE "ALL"
;;            IDLE
        @6, dlg.col
        height 11
        width msgwid

        style attribute msgcolor
        @1,1  ?? format( fmtw, iif( isassigned( ww[1] ), ww[1], "" ) )
        @2,1  ?? format( fmtw, iif( isassigned( ww[2] ), ww[2], "" ) )
        @3,1  ?? format( fmtw, iif( isassigned( ww[3] ), ww[3], "" ) )
        @4,1  ?? format( fmtw, iif( isassigned( ww[4] ), ww[4], "" ) )
        @5,1  ?? format( fmtw, iif( isassigned( ww[5] ), ww[5], "" ) )
        style

        pushbutton
            @7,int( (msgwid - 17 ) / 2 )
            width 14
            but1
            cancel
            default
            value but1
            tag but1
            to butval

    enddialog

endswitch

endproc
wproc( "UTLX_Ok_Box" )

;================================================== UTLX_Parse_Path
proc UTLX_Parse_Path
; Split path into its component parts.
;
; Return Value:
; None.
; The components in the path are in the dynarray parsed.path.
(
    path.to.parse     ; {I} Path to parse into component parts.
  , parsed.path       ; {O} Dynarray that is filled with path
                      ;     components by this proc.
                      ;     Sequentially indexed starting at 1.
)
private x, z

; Make sure path has "\" at end
if substr( path.to.parse, len( path.to.parse ), 1 ) <> "\\" then
    path.to.parse = path.to.parse + "\\"
endif

x = 0
while match( path.to.parse, "..\\..", z, path.to.parse )
    x = x + 1
    parsed.path[ x ] = z
endwhile

endproc
wproc( "UTLX_Parse_Path" )

;================================================== UTLX_Concat_Path
proc UTLX_Concat_Path
; Concatenate components into a path.
; This is the inverse function for UTLX_Parse_Path.
;
; Return Value:
; The concatenated path with "\" at the end.
(
    path.part        ; {I} Dynarray containing path components.
                     ;     Sequentially indexed starting at 1.
  , path.lvl         ; {I} Number of path levels to concatenate.
                     ;     Should be <= dynarraysize( path.part ).
)
private x, z

z = ""
for x from 1 to min( path.lvl, dynarraysize( path.part ) )
    z = z + path.part[ x ] + "\\"
endfor

return z

endproc
wproc( "UTLX_Concat_Path" )

;============================================== UTLX_Please_Wait
proc UTLX_Please_Wait
; Displays a "Please Wait .." message.
;
; The message is in a floating canvas and is centered on the screen.
; Each line in the message will be centered.
;
; To remove the message use proc UTLX_Close_Canvas.
;
; Return Value:
; Window handle of message window.
; The message window is left on the screen.
(
    msg        ; {I} Message to be displayed.
               ;     The message will be word wrapped. It must not
               ;     exceed screen height minus 4 lines.
  , hgt        ; {I} Height of message window.
               ;     If blank or zero, the height will be
               ;     calculated automatically (height of msg).
               ;     Use a negative number to specify the
               ;     top and bottom margins. E.g., -3 sets
               ;     the top and bottom margins to 3.
  , wid        ; {I} Width of message window.
               ;     If blank or zero, the width will be
               ;     calculated automatically (width of msg).
               ;     Use a negative number to specify the
               ;     left and right margins. E.g., -5 sets
               ;     the left and right margins to 5
)
private x


return UTLX_Open_Msg_Window( msg, true
    , "Please Wait ..", true, hgt, wid, 0 )

endproc
wproc( "UTLX_Please_Wait" )

;======================================================= UTLX_Open_Msg_Window
proc UTLX_Open_Msg_Window
; Create a floating window with a message.
;
; The window will centered on the screen.
;
; This is a non-modal message. The message is displayed and
; execution continues.
;
; The window will remain until it is closed (via a call to
; the proc UTLX_Close_Canvas). It can be used as the
; target for canvas writing commands.
;
; The original canvas is restored upon exit. If you wish to use
; the message window for canvas commands, you must issue a
; SETCANVAS command.
;
; Return Value:
; Window handle.
; The floating window with the message is left on the screen.
(
    msgtxt        ; {I} Message to display.
                  ;     The message will be word wrapped and can be
                  ;     up to hgt - 2 lines long. Excess lines will
                  ;     be ignored.
                  ;     The 2 lines are used for a blank line at
                  ;     the top and bottom of the window.
  , ctrtxt        ; {I} If TRUE, then each line of the message
                  ;     is centered in the window.
  , titl          ; {I} Title for window.
                  ;     If the window is framed, the title is
                  ;     in the top frame. (As usual)
                  ;     If the window is not framed, the title
                  ;     is in the top row of the window. The
                  ;     row has the color attribute of an active
                  ;     frame. The window height is automatically
                  ;     increased by 1 to allow room for the title.
  , fram          ; {I} If true, the window is framed and not
                  ;     shadowed.
                  ;     If false, the window is shadowed and not
                  ;     framed.
                  ;     If the window is framed, the height and
                  ;     width are automatically increased by two.
  , hgt           ; {I} Window height.
                  ;     If blank or zero, then the window will
                  ;     be n + 2 lines high, where n is the number
                  ;     of lines in msg.
                  ;     The maximum height is screen height - 4.
                  ;     The minimum height is 3.
                  ;     Use a negative number to specify the
                  ;     top and bottom margins. E.g., -3 sets
                  ;     the top and bottom margins to 3.
  , wid           ; {I} Window width.
                  ;     If blank or zero, then the window will
                  ;     be n + 2 columns wide, where n is the width
                  ;     of the longest line in msg.
                  ;     The maximum width is screen width - 4.
                  ;     The minimum width is 3.
                  ;     Use a negative number to specify the
                  ;     left and right margins. E.g., -5 sets
                  ;     the left and right margins to 5
  , styl          ; {I} Text style.
                  ;     If blank or zero, the default style will
                  ;     be used. The default style is the same as
                  ;     the color used for a label's hotkey in a
                  ;     dialog box, syscolor( 1039 )
                  ;     If the window is framed, then the frame will
                  ;     have the same background color as the text.
)
private x, y, z
    , ww               ; dynarray of word wrapped text
    , scr.hgt          ; screen height
    , scr.wid          ; screen width
    , winhgt           ; window height
    , winwid           ; window width
    , winhand          ; window handle
    , orig.canv        ; original canvas handle
    , fmtw             ; string for text format command
    , fmtw.title       ; string for title format command
    , fake.title       ; true if window is unframed and has a title
    , lr.marg          ; left/right margins
    , tb.marg          ; top/bottom margins

sysinfo to x
scr.hgt = x[ "screenheight" ]
scr.wid = x[ "screenwidth" ]
release vars x

if isblank( styl ) or styl = 0 then
    styl = syscolor( 1039 )
endif

lr.marg = 1
tb.marg = 1

if not isblank( hgt ) and hgt < 0 then
    tb.marg = -hgt
    hgt = 0        ; auto calc height
endif

if not isblank( wid ) and wid < 0 then
    lr.marg = -wid
    wid = 0        ; auto calc width
endif

if isblank( hgt ) or hgt < 1 then
    winhgt = scr.hgt - 4
    hgt = 0
else
    winhgt = max( hgt, 3 )
endif

if isblank( wid ) or wid < 1 then
    winwid = scr.wid - 4
    wid = 0
else
    winwid = max( wid, 3 )
endif

dynarray ww[]
ww[ "string" ] = msgtxt
ww[ "width" ] = winwid - 2

UTLX_Word_Wrap( ww )

if hgt = 0 then
    while ww[ "lines" ] + 2 * tb.marg > scr.hgt - 4
        if tb.marg > 1 then
            tb.marg = tb.marg - 1
        else
            quitloop
        endif
    endwhile
    winhgt = min( ww[ "lines" ] + 2 * tb.marg, scr.hgt - 4 )
endif
if wid = 0 then
    while ww[ "longest" ] + 2 * lr.marg > scr.wid - 4
        if lr.marg > 1 then
            lr.marg = lr.marg - 1
        else
            quitloop
        endif
    endwhile
    winwid = min( ww[ "longest" ] + 2 * lr.marg, scr.wid - 4 )
endif

if wid = 0 and not isblank( titl ) then
    if fram then
        ; framed window is 10 wider than the title.
        x = min( len( titl ) + 10, scr.wid - 4 )
    else
        ; unframed window is 2 wider than title.
        x = min( len( titl ) + 2, scr.wid - 4 )
    endif
    winwid = max( winwid, x )
endif

fake.title = not ( fram or isblank( titl ) )

fmtw = "w" + strval( winwid - 2 )
if ctrtxt then
    fmtw = fmtw + ",ac"
endif

fmtw.title = "w" + strval( winwid ) + ",AC"

z = iif( fram, 2, 0 )  ; increment for frame
y = iif( fake.title, 1, 0 ) ; room for title in text

dynarray x[]
x[ "canclose" ] = false
x[ "canmaximize" ] = false
x[ "canmove" ] = false
x[ "canresize" ] = false
x[ "canvasheight"] = winhgt + y
x[ "canvaswidth" ] = winwid
x[ "floating" ] = true
x[ "hasframe" ] = fram
x[ "hasshadow" ] = not fram
x[ "height" ] = winhgt + z + y ; allow for frame or fake title
x[ "maximized" ] = false
x[ "origincol" ] = int( ( scr.wid - winwid - z ) / 2 )  ; bias left
x[ "originrow" ] = int( ( scr.hgt - winhgt - z ) / 2 )  ; bias up
x[ "style" ] = styl
x[ "title" ] = titl        ; ignored if window not framed
x[ "width" ] = winwid + z  ; allow for frame

window create attributes x to winhand
release vars x
if fram then
    dynarray x[]
    x[ 0 ] = 0  + 16 * int( styl / 16 )  ; inactive frame
    x[ 1 ] = 15 + 16 * int( styl / 16 )  ; active frame
    x[ 2 ] = 15 + 16 * int( styl / 16 )  ; frame icons
    window setcolors winhand from x
    release vars x
endif

orig.canv = getcanvas()
setcanvas winhand

if fake.title then
    @ 0, 0 ?? format( fmtw.title, titl )
    z = 15 + 16 * int( styl / 16 )  ; title color
    paintcanvas attribute z 0, 0, 0, winwid - 1
    z = tb.marg
else
    z = tb.marg - 1
endif
for x from 1 to min( ww[ "lines" ], winhgt - 2 )
    @ x + z, 1
    ?? format( fmtw, ww[ x ] )
endfor

if iswindow( orig.canv ) then
    setcanvas orig.canv
else
    setcanvas default
endif

return winhand

endproc
wproc( "UTLX_Open_Msg_Window" )

;======================================================= UTLX_Close_Canvas
proc UTLX_Close_Canvas
; Close any window (Intended for canvas windows).
;
; The current canvas is left as-is, set to the default canvas,
; or set to the canvas passed in new.canv.
;
;
; Return Value:
; None.
; The current canvas may be set.
(
    canv          ; {I} This is the window handle of the window
                  ;     that is to be closed.
  , new.canv      ; {I} The window handle for the new current canvas.
                  ;     If not a valid window handle, then the
                  ;     default canvas will become the current
                  ;     canvas if the current canvas is invalid.
)
private x

if iswindow( canv ) then
    x = getwindow()
    window select canv
    window close
    if type( new.canv ) <> "N" then
        new.canv = 0
    endif

    if iswindow( new.canv ) then
        setcanvas new.canv
    else
        if not iswindow( getcanvas() ) then
            setcanvas default
        endif
    endif

    if iswindow( x ) then
        window select x
    endif
endif

endproc
wproc( "UTLX_Close_Canvas" )

;================================================== UTLX_Box_DlgProc
proc UTLX_Box_DlgProc
; Dialog proc for UTLX_Ok_Box, UTLX_Ok_Cancel_Box, and
; UTLX_Cancel_Ok_Box.
;
; The proc is set up to handle all events but the dialog boxes
; only trap the open event.
;
; Return Value:
; True if the event is to be processed.
; False if the event is to be ignored.
(
    dlg.trig     ; "Event" or one of the following triggers:
                 ; "Update", "Depart", "Select"
                 ; "Accept", "Cancel", "Arrive"
                 ; "Open", "Close"
  , dlg.tag      ; The tag for the active control element
  , dlg.evt      ; Dynarray for event.
                 ; If "Update", value of active control element.
                 ; If "Depart", tag that you will navigate to.
                 ; Null string if any other trigger.
  , dlg.cb.lbl   ; Null string if event.
                 ; If "Update" trigger, label string of checkbox or
                 ; null string.
                 ; Null string if any other trigger.
)
private x
switch
case dlg.trig = "EVENT" :
    switch
    case dlg.evt[ "type" ] = "IDLE"    : return false
    case dlg.evt[ "type" ] = "KEY"     : return UTLX_Box_DlgEvt_Keys()
    case dlg.evt[ "type" ] = "MOUSE"   : return true
    case dlg.evt[ "type" ] = "MESSAGE" : return true
    endswitch
case dlg.trig = "UPDATE" : return true
case dlg.trig = "DEPART" : return true
case dlg.trig = "SELECT" : return true
case dlg.trig = "ACCEPT" : return true
case dlg.trig = "ARRIVE" : return true
case dlg.trig = "OPEN"   : return UTLX_Box_DlgTrg_Open()
case dlg.trig = "CANCEL" : return true
case dlg.trig = "CLOSE"  : return true
endswitch
return true

endproc
wproc( "UTLX_Box_DlgProc" )

;============================================== UTLX_Box_DlgTrg_Open
proc UTLX_Box_DlgTrg_Open
; Handles OPEN trigger for the "Box" dialogs.
;
; It sets the picklist colors to msgcolor parameter (from the
; proc that does the ShowDialog).
;
; Return Value:
; True or false.
(
)
private x
    , dlghand
    , f.mcolor
    , b.mcolor
    , inv.mcolor

window handle dialog to dlghand

; Need to invert msgcolor for attribute 26 (the selected item
; when the picklist is active.

b.mcolor = int( msgcolor / 16 )
f.mcolor = mod( msgcolor, 16 )
if f.mcolor > 8 then
    f.mcolor = f.mcolor - 8
    b.mcolor = b.mcolor + 8
endif
inv.mcolor = ( 16 * f.mcolor ) + b.mcolor

dynarray x[]
x[ 25 ] = msgcolor
x[ 26 ] = inv.mcolor
x[ 27 ] = msgcolor
window setcolors dlghand from x

return true

endproc
wproc( "UTLX_Box_DlgTrg_Open" )

;============================================== UTLX_Box_DlgEvt_Keys
proc UTLX_Box_DlgEvt_Keys
; Handles KEY events for the message box dialogs.
;
; Return Value:
; True or false.
(
)
private x, y, z

; Make F2 push the Accept button
if dlg.evt[ "keycode" ] = asc( "F2" ) then
    release vars dlg.evt[ "scancode" ]
    dlg.evt[ "keycode" ] = asc( "Enter" )
    selectcontrol accept.butval
    return true
endif

; It's not really necessary, but we'll kill the Ctrl-O key.
; We allow Alt-O because it may be tied to the Ok button for the dialog.
if dlg.evt[ "keycode" ] = 15 then
    beep
    return false
endif

; Show help if the user presses F1
if dlg.evt[ "keycode" ] = asc( "F1" ) then
    if not isblank( dlg.help ) then
        execproc dlg.help
    endif
    return false
endif

return true

endproc
wproc( "UTLX_Box_DlgEvt_Keys" )

;================================================== UTLX_DriveStatus
proc UTLX_DriveStatus
; Check the status of a drive via a DRIVESTATUS function.
;
; Since DRIVESTATUS can take a long time to execute, a
; "Please Wait" message is displayed. Since this message will
; barely flicker for fast drives, a slight delay is added
; to make the message readable.

; Since this function may be called several times during an
; interaction, the message is automatically suppressed after
; the first call.
; To enable the message, the caller must set the global var
; utlx.drivestatus.msg to TRUE for the first call in a series.
;
; Return Value:
; True or false as returned by DRIVESTATUS
(
    drv        ; {I} The drive to check. Only the first byte
               ;     is passed to DRIVESTATUS.
)
private x
    , msghand

drv = upper( substr( drv, 1, 1 ) )

if not isassigned( utlx.drivestatus.msg ) then
    utlx.drivestatus.msg = true
endif

if utlx.drivestatus.msg then
    utlx.drivestatus.msg = false
    msghand = UTLX_Please_Wait( "Checking status of drive \""
        + drv + ":\".", 0, -8 )
    sleep 500
else
    msghand = 0
endif

x = drivestatus( drv )

UTLX_Close_Canvas( msghand, 0 )

return x

endproc
wproc( "UTLX_DriveStatus" )

;=========================================================== UTLX_Single_Beep
proc UTLX_Single_Beep
; Go beep
;
; Return Value:
; None.
(
)
private w, x, y, z
    , lf, hf, sf, dur

beep

endproc
wproc( "UTLX_Single_Beep" )

;=========================================================== UTLX_Double_Beep
proc UTLX_Double_Beep
; Go beep beep
;
; Return Value:
; None.
(
)
private w, x, y, z
    , lf, hf, sf, dur

beep
sleep 10
beep

endproc
wproc( "UTLX_Double_Beep" )

;=========================================================== UTLX_Error_Beep
proc UTLX_Error_Beep
; Beep used to get user's attention when an error occurs.
;
; Return Value:
; None.
(
)
private w, x, y, z
    , lf, hf, sf, dur

lf = 100
hf = 400
sf = 100
dur = iif(utlx.turbopal, 2, 1 )
for x from lf to hf step sf
    sound x dur
endfor
for x from hf to lf step -sf
    sound x dur
endfor

endproc
wproc( "UTLX_Error_Beep" )

;=========================================================== UTLX_Crash_Beep
proc UTLX_Crash_Beep
; Whistle and fart for internal warning/error.
;
; Return Value:
; None.
(
)
private w, x, y, z
    , lf, hf, sf, dur

lf = 100
hf = 1000
sf = 50

dur = iif(utlx.turbopal, 2, 1 )

for x from hf to lf step -sf
    sound x dur
    sound x - int( sf / 2 ) dur
endfor

dur = iif(utlx.turbopal, 5, 10 )
lf = 25

for x from 1 to 15
    sound lf - x dur
    sound lf - x - 1 dur
endfor

endproc
wproc( "UTLX_Crash_Beep" )

;===================================================== UTLX_Word_Wrap
proc UTLX_Word_Wrap
; Word wraps a string.
;
; Return Value:
; None.
;
; This proc updates the dynarray parameter as follows:
;
;   str["lines"] is set to the number of lines.
;
;   str["longest"] is set to the length of the longest line after
;   wrapping.  This value will be less than or equal to str["width"].
;
;   str[n] is set to the contents of line n,
;   where 1 <= n <= str["lines"].
(
    str     ; {I,O} Dynarray that contains the string to be word
            ;       wrapped in the element str["string"] and the
            ;       wrapping width in str["width"].
            ;       Wrapping width must be between 1 and 255.
)
private x, y
    , z            ; string from str["string"]
    , w            ; wrap width from str["width"]
    , s            ; current start pos in z
    , e            ; current end pos in z
    , sz        ; length of substring in z
    , lz        ; length of z

; Check parameters in str

if isassigned( str["string"] ) then
    z = str["string"]
else
    z = ""
endif

if isassigned( str["width"] ) then
    w = numval( str["width"] )
    if lower( strval( w ) ) = "error" then
        w = 20   ; default width
        z = z + "\n*** Invalid width passed ***"
    endif
    if w < 1 or w > 255 then
        w = 20   ; default width
        z = z + "\n*** Invalid width passed ***"
    endif
else
    w = 20       ; default width
    z = z + "\n*** No width passed ***"
endif

; Word wrap lines

str["lines"] = 0
str["longest"] = 0
s = 1
e = 1
lz = len( z )

while s <= lz
    e = searchfrom( "\n", z, s )
    if e > 0 and e - s <= w then
        sz = e - s   ; don't include \n
    else
        ; There is no \n in the w bytes starting at s.
        ; Look for rightmost blank or dash in string of width w.
        sz = 0
        e = s + w - 1
        while e > s and e < lz
            if substr( z, e, 1 ) = " " then
                sz = e - s  ; don't include trailing blank
                quitloop
            endif
            if substr( z, e, 1 ) = "-" then
                sz = e - s + 1 ; include trailing dash
                quitloop
            endif
            e = e - 1
        endwhile
        if sz = 0 then
            ; no natural breaks in line or at end of string - use
            ; maximum width
            e = s + w -1
            e = min( e, lz )
            sz = e - s + 1
        endif
    endif
    ; trim trailing blanks from line (caller may want to center lines)
    y = substr( z, s, sz )
    while match( y, ".. ", y ) endwhile

    str["lines"] = str["lines"] + 1
    str[ str["lines"] ] = y
    if len( y ) > str["longest"] then
        str["longest"] = len( y )
    endif
    s = e + 1
endwhile

endproc
wproc( "UTLX_Word_Wrap" )

